#ifndef CODEWAR_MAIN_HEADER
#define CODEWAR_MAIN_HEADER

//
// This is the main header file for the Code War project
// It is used both by the server and the C++ client
// -- by Gabriel Fleseriu
// First version: 01/19/2003
//
#include <vector>
#include <iterator>
#include <iostream>
namespace codewar{

#pragma pack(push,codewar)
#pragma pack(1)

//
// message id constants
//

const short msgid_base = 0;
const short msgid_creature_stats    = msgid_base + 0;
const short msgid_server_handshake  = msgid_base + 1;
const short msgid_client_handshake  = msgid_base + 2;
const short msgid_creature_feedback = msgid_base + 3;
const short msgid_creature_action   = msgid_base + 4;

//
// message header
//

struct msg_head
{
    short size;
    short id;
};


//
// server handshake
// when a client has connected, the server gives that client a species id
// and tells it how many creatures it will initially start with
//

struct server_handshake
{
    server_handshake(short sid = -1, short nc = -1)
    {
        species_id = sid;
        no_of_creatures = nc;
    }
    short species_id;
    short no_of_creatures;
};

struct server_handshake_message
{
    server_handshake_message(short sid = -1, short nc = -1)
    {
        head.size = sizeof(server_handshake_message);
        head.id = msgid_server_handshake;
        handshake.species_id = sid;
        handshake.no_of_creatures = nc;
    }
    server_handshake_message(const server_handshake &h)
    {
        head.size = sizeof(server_handshake_message);
        head.id = msgid_server_handshake;
        handshake = h;
    }
    msg_head head;
    server_handshake handshake;
};

//
// client handshake
// after receiving the server handshake the client transmits its
// species id back and a name for the species (max 31 chars)
// this way the server will be able to write the mapping 
// species_id <--> client to the log file
//

struct client_handshake
{
    client_handshake(short sid = -1, char *sname = 0)
    {
        species_id = sid;
        if(sname){
            strncpy(species_name, sname, 32);
        }
        else{
            species_name[0] = 0;
        }
    }
    short species_id;
    char species_name[32];
};

struct client_handshake_message
{
    msg_head head;
    client_handshake handshake;
    client_handshake_message(short sid=-1, char *sname = 0)
    {
        head.size = sizeof(client_handshake_message);
        head.id = msgid_client_handshake;
        handshake.species_id = sid;
        if(sname){
            strncpy(handshake.species_name, sname, 32);
        }
        else{
            handshake.species_name[0] = 0;
        }
    }
    client_handshake_message(const client_handshake & h)
    {
        head.size = sizeof(client_handshake_message);
        head.id = msgid_client_handshake;
        handshake = h;
    }
};

//
// creature stats
// after a completed handshake, the client sends its creature stats to the server
//

struct creature_stats
{
    creature_stats(short a = 10, short d = 10, short f = 10, short l = 10)
    {
        attack = a;
        defense = d;
        food_consumption = f;
        max_life = l;
    }
    short attack;
    short defense;
    short food_consumption;
    short max_life;
};

struct creature_stats_message
{
    msg_head head;
    creature_stats stats;
    creature_stats_message(short a = 10, short d = 10, short f = 10, short l = 10)
    {
        head.size = sizeof(creature_stats_message);
        head.id = msgid_creature_stats;
        stats.attack = a;
        stats.defense = d;
        stats.food_consumption = f;
        stats.max_life = l;
    }
    creature_stats_message(const creature_stats & s)
    {
        head.size = sizeof(creature_stats_message);
        head.id = msgid_creature_stats;
        stats = s;
    }
};

//
// cell description constants
//

const short cell_void           = 0; // outside the arena
const short cell_empty          = cell_void + 1;
const short cell_occluded       = cell_void + 2;
const short cell_food           = cell_void + 3;
const short cell_life           = cell_void + 4;
const short cell_wall           = cell_void + 5;
const short cell_creature       = cell_void + 6;

//
// direction constants
//

const short north = 0;
const short west  = 1;
const short south = 2;
const short east  = 3;

//
// creature action constants
//

const short idle        = 0; // no decision
const short rest        = idle + 1;
const short move        = idle + 2;
const short turn90      = idle + 3; // in trigonometrical sense, i.e. N->W->S->E->N
const short turn180     = idle + 4;
const short turn270     = idle + 5;
const short feed        = idle + 6;
const short revive      = idle + 7; // replenish life on a life cell

//
// basic creature description
//

struct basic_creature
{
    short species;
    short id;
    short x;
    short y;
    short facing;
    short life;
    short food;
    void push_to_vector(std::vector<short> &v) const
    {
        v.push_back(species);
        v.push_back(id);
        v.push_back(x);
        v.push_back(y);
        v.push_back(facing);
        v.push_back(life);
        v.push_back(food);
    }
    bool pull_from_vector(const std::vector<short> &v, std::vector<short>::const_iterator &it)
    {
        if(it == v.end()) return false;
        species = *it++;if(it == v.end()) return false;
        id = *it;++it;if(it == v.end()) return false;
        x=*it;++it;if(it == v.end()) return false;
        y=*it;++it;if(it == v.end()) return false;
        facing=*it;++it;if(it == v.end()) return false;
        life=*it;++it;if(it == v.end()) return false;
        food=*it;++it;if(it == v.end()) return false;
        return true;
    }
};

//
// sighted creature information
// a creature will receive this info about another creature it sees
//

struct sighted_creature_info
{
    sighted_creature_info():species(-1), id(-1), life(0), facing(north){}
    short species;
    short id;
    short life; // in %
    short facing;
};

//
// attacked info
// if a creature gets attacked, it will receive this info
// 
struct attacked_by_info
{
    attacked_by_info(short s = -1, short i = -1, short d = 0, short f = north):
        species(s), id(i), damage(d), from(f){}
    short species; // -1 if no attack: ignore the rest
    short id;
    short damage;
    short from; // N, W, S or E
    void push_to_vector(std::vector<short> &v) const
    {
        v.push_back(species);
        v.push_back(id);
        v.push_back(damage);
        v.push_back(from);
    }
    bool pull_from_vector(const std::vector<short> &v, std::vector<short>::const_iterator &it)
    {
        if(it == v.end()) return false;
        species = *it++;if(it == v.end()) return false;
        id = *it;++it;if(it == v.end()) return false;
        damage=*it;++it;if(it == v.end()) return false;
        from=*it;++it;if(it == v.end()) return false;
        return true;
    }
};

//
// action result constants
// don't know if we really need them -- the result of an action actually can be deduced fron the state difference
//
const short result_ok = 0;
const short result_attack = 1; // attack succeeded
const short result_error = 2;  // e.g. move into wall, harvest empty cell etc.

//
// cell numbering schema
//
/*
1 2 3 4 5   5 x x x x   x x x x x   x x x x 1
x 6 7 8 x   4 8 x x x   x x x x x   x x x 6 2
x x ^ x x   3 7 < x x   x x V x x   x x > 7 3
x x x x x   2 6 x x x   x 8 7 6 x   x x x 8 4
x x x x x   1 x x x x   5 4 3 2 1   x x x x 5
*/

//
// creature feedback info
// this is the info each client gets for each of its creatures at the begin of a turn
// this structure is informative only, that is why it is excluded from compilation
//
#if 0
struct creature_feedback_message_draft
{
    msg_head head;
    basic_creature creature;
    attacked_by_info attacked;
    short last_turn_result;
    vision_info vision;  // this has variable length
};
#endif


//
// basic vision info
// the vision information has variable length. it is basically an array of shorts
// containing cell_... constants in the order outlined in the cell numbering schema
// however, if a field contains a creature, the contents of the sighted_creature_info
// is appended directly after the cell_creature element
//

struct basic_vision_info
{
    std::vector<short> data;
    void push_back(short s){data.push_back(s);}
    void push_back(const sighted_creature_info& c)
    {
        data.push_back(c.species);
        data.push_back(c.id);
        data.push_back(c.life);
        data.push_back(c.facing);
    }
    int size() const {return data.size();}
};

//
// creature feedback message
// this is used by the server to send the message
// the client has to parse it
//

struct creature_feedback_message
{
    std::vector<short> data;
    creature_feedback_message(const basic_creature & creature,
                              const attacked_by_info & attacked,
                              const short last_turn_result, 
                              const basic_vision_info & vision)
    {
        data.push_back(0); // == head.size
        data.push_back(msgid_creature_feedback); // == head.id
        creature.push_to_vector(data);
        attacked.push_to_vector(data);
        data.push_back(last_turn_result);
        std::copy(vision.data.begin(), vision.data.end(), std::back_inserter(data));
        data[0] = data.size()*sizeof(short);
    }
    void *message(){return &data[0];}
    int size(){return data[0];}
};

//
// client vision info
// this is a representation of the vision info for the client
// a more elaborated client may choose another represenation
//


struct client_vision_info
{
    struct client_cell_info{
        client_cell_info():cell_contents(cell_void){}
        short cell_contents;
        sighted_creature_info creature; //not a pointer by purpose; wastes space, but saves allocation head aches
    };
    client_cell_info cells[8];
    bool parse(const std::vector<short> &v, std::vector<short>::const_iterator &it)
    {
        // parse the vector from it to v.end() 
        // very crude implementation
        if(it == v.end()) return false;
        int i = 0;
        while(it != v.end()){
            switch(*it){
            case cell_creature:
                cells[i].cell_contents = *it;++it;
                if(it == v.end()) return false;
                cells[i].creature.species = *it;++it;
                if(it == v.end()) return false;
                cells[i].creature.id = *it;++it;
                if(it == v.end()) return false;
                cells[i].creature.life= *it;++it;
                if(it == v.end()) return false;
                cells[i].creature.facing = *it;++it;
                ++i;
                break;
            default:
                cells[i].cell_contents = *it;++it;
                ++i;
                break;
            }
        }
        return true;
    }
};

//
// client feedback info
// this is a representation of the feedback info for the client
// a more elaborated client may choose another represenation
//

struct client_feedback_info
{
    basic_creature creature;
    attacked_by_info attacked;
    short last_turn_result;
    client_vision_info vision;
    bool parse(const std::vector<short> &v)
    {
        std::vector<short>::const_iterator it = v.begin();
        ++it; ++it; // jump over msg_head
        if(!creature.pull_from_vector(v, it)) return false;
        if(!attacked.pull_from_vector(v, it)) return false;
        if(it == v.end()) return false;
        last_turn_result = *it;++it;
        if(it == v.end()) return false;
        if(!vision.parse(v, it)) return false;
        return true;
    }
};

//
// client feedback message
// this is used by the client to receive the feedback message over the socket
//

struct client_feedback_message
{
    msg_head head;
    std::vector<short> data;
    void prepare(){data.resize(head.size/sizeof(short));data[0] = head.size;}
    void *message(){return &data[0];}
    int size(){return data[0];}
};

//
// creature action message
// sent by the client for each creature to tell server what next action is
//

struct creature_action_message
{
    creature_action_message(short s = -1, short c = -1, short a = idle)
    {
        head.id = msgid_creature_action;
        head.size = sizeof(creature_action_message);
        species = s;
        creature = c;
        action = a;
    }
    msg_head head;
    short species;
    short creature;
    short action;
};

//
// war parameters -- to avoid the use of "magic numbers"
//
struct war_parameters_struct
{
    war_parameters_struct(short creatures = 10, int turns = 1000, unsigned int s = 100):
no_of_creatures(creatures), no_of_turns(turns), starting_food(s){};
    short no_of_creatures;
    int no_of_turns;
    short starting_food;
};

const unsigned int connection_port = 1234;
#pragma pack(pop, codewar)
} //namespace codewar
#endif //CODEWAR_MAIN_HEADER